import logging
from typing import Any

from web3 import Web3
from .ether import EtherNode
from web3.middleware import geth_poa_middleware
from eth_account import Account

logger = logging.getLogger('nodes.bsc')


class BscNode(EtherNode):
    transfer_sigs = {
        'a9059cbb': ['transfer', 'address', 'uint256']
    }
    contract_symbols = {
        Web3.toChecksumAddress('0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82'): 'CAKE',
        Web3.toChecksumAddress('0xba2ae424d960c26247dd6c32edc70b295c744c43'): 'DOGE',
        Web3.toChecksumAddress('0x8fF795a6F4D97E7887C79beA79aba5cc76444aDf'): 'BCH',
        Web3.toChecksumAddress('0x1D2F0da169ceB9fC7B3144628dB156f3F6c60dBE'): 'XRP',
        Web3.toChecksumAddress('0x43a172c44dC55c2B45BF9436cF672850FC8bA046'): 'METAWAR',
        Web3.toChecksumAddress('0xf8a0bf9cf54bb92f17374d9e9a321e6a111a51bd'): 'LINK',
        Web3.toChecksumAddress('0x352Cb5E19b12FC216548a2677bD0fce83BaE434B'): 'BTT',
    }
    native_token = 'BNB'

    def __init__(self, uri: str, chain_id: int):
        super(BscNode, self).__init__(uri=uri, chain_id=chain_id)

        # inject the poa compatibility middleware to the innermost layer (0th layer)
        # @TODO: the following injection must happen before all other middleware_onion calls
        # ref: https://www.reddit.com/r/ethdev/comments/pbb536/estimating_a_gas_price_based_on_a_target_time_of/
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        logger.info('Injected geth_poa_middleware layer 0')

    def sign_function(self, function_call: Any, private_key: str):
        account = Account.from_key(private_key)

        # Prepare transaction
        _nonce = self._w3.eth.get_transaction_count(account.address)
        tx = {
            'chainId': self._chain_id,
            'nonce': _nonce,
            'from': account.address,
        }

        # Build transaction with function_call
        tx = function_call.build_transaction(tx)

        # Gas calculation
        tx['gas'] = self._w3.eth.estimate_gas(tx)
        tx['gasPrice'] = self._w3.eth.gas_price

        # sign the transaction
        signed_tx = self._w3.eth.account.sign_transaction(tx, private_key)

        return signed_tx['hash'], signed_tx

    def sign_transact(
            self, private_key: str, to_address: str, amount: int, fees_included: bool = False) -> tuple[bytearray, Any]:
        account = Account.from_key(private_key)

        # Prepare transaction
        _nonce = self._w3.eth.get_transaction_count(account.address)
        tx = {
            'chainId': self._chain_id,
            'nonce': _nonce,
            'from': account.address,
            'to': Web3.toChecksumAddress(to_address),
        }

        # Gas calculation
        tx['gas'] = self._w3.eth.estimate_gas(tx)
        tx['gasPrice'] = self._w3.eth.gas_price

        if fees_included:
            # Include the fees in the amount
            tx['value'] = amount - (tx['gas'] * tx['gasPrice'])
        else:
            tx['value'] = amount

        # sign the transaction
        signed_tx = self._w3.eth.account.sign_transaction(tx, private_key)

        return signed_tx['hash'], signed_tx



from dataclasses import dataclass
from dataclasses import field


@dataclass
class Transfer(object):
    hash: bytearray = field(metadata={"required": True})
    method_id: str = field(metadata={"required": True})

    status: bool = field(metadata={"required": True})
    sender: str = field(metadata={"required": True})
    receiver: str = field(metadata={"required": True})
    symbol: str = field(metadata={"required": True})
    is_token: bool = field(metadata={"required": True})
    value: int = field(metadata={"required": True})

    trx_index: int = None
    block_number: int = None
    block_hash: bytearray = None



from dataclasses import dataclass
from dataclasses import field


@dataclass
class Block(object):
    number: int = field(metadata={"required": True})
    hash: bytearray = field(metadata={"required": True})

    transaction_count: int = field(metadata={"required": True})



import logging
from typing import Sequence, Union, Any

from .interface import INode
import secrets
from cachetools import cached, TTLCache
from eth_account import Account
from eth_abi import abi
from web3 import Web3
from web3.exceptions import TransactionNotFound
from web3.providers import HTTPProvider
from web3.gas_strategies.time_based import construct_time_based_gas_price_strategy
from web3 import middleware

from hexbytes import HexBytes

from .types import Block, Transfer, Receipt
from automapper import mapper

logger = logging.getLogger('nodes.ether')

is_contract_cache = TTLCache(maxsize=500, ttl=3600)


class EtherNode(INode):
    transfer_sigs = {
        'a9059cbb': ['transfer', 'address', 'uint256']
    }
    contract_symbols = {
        Web3.toChecksumAddress('0x0d8775f648430679a709e98d2b0cb6250d2887ef'): 'BAT',
        Web3.toChecksumAddress('0x95ad61b0a150d79219dcf64e1e6cc01f0b64c4ce'): 'SHIB',
        Web3.toChecksumAddress('0xdac17f958d2ee523a2206206994597c13d831ec7'): 'USDT',
        Web3.toChecksumAddress('0x2af5d2ad76741191d15dfe7bf6ac92d4bd912ca3'): 'LEO',
        Web3.toChecksumAddress('0xa0b73e1ff0b80914ab6fe0444e65848c4c34450b'): 'CRO',
        Web3.toChecksumAddress('0x6f259637dcd74c767781e37bc6133cd6a68aa161'): 'HT',
        Web3.toChecksumAddress('0x3506424f91fd33084466f402d5d97f05f8e3b4af'): 'CHZ',
        Web3.toChecksumAddress('0x0f5d2fb29fb7d3cfee444a200298f468908cc942'): 'MANA',
        Web3.toChecksumAddress('0xb62132e35a6c13ee1ee0f84dc5d40bad8d815206'): 'NEXO',
        Web3.toChecksumAddress('0x514910771af9ca656af840dff83e8264ecf986ca'): 'LINK',
        Web3.toChecksumAddress('0xc669928185dbce49d2230cc9b0979be6dc797957'): 'BTT',
    }
    native_token = 'ETH'
    default_abi = {
        'transfer': [
            {"constant": False, "inputs": [{"name": "_to", "type": "address"}, {"name": "_value", "type": "uint256"}],
             "name": "transfer", "outputs": [{"name": "success", "type": "bool"}], "payable": False,
             "stateMutability": "nonpayable", "type": "function"}],
        'balanceOf': [
            {"constant": True, "inputs": [{"name": "_owner", "type": "address"}], "name": "balanceOf",
             "outputs": [{"name": "balance", "type": "uint256"}], "payable": False, "stateMutability": "view",
             "type": "function"}]
    }
    miner_fee_percent = 0.02

    def __init__(self, uri: str, chain_id: int, max_wait_seconds: int = 60, sample_size: int = 3,
                 probability: int = 95):
        super(EtherNode, self).__init__(uri=uri)
        logger.debug('Instantiating EtherNode (chainId: %s) with uri: %s', chain_id, uri)
        self._w3 = Web3(HTTPProvider(endpoint_uri=self._uri))

        # Set gas pricing strategy
        self._w3.eth.set_gas_price_strategy(
            construct_time_based_gas_price_strategy(
                max_wait_seconds=max_wait_seconds, sample_size=sample_size, probability=probability))

        # Use cacheing
        self._w3.middleware_onion.add(middleware.time_based_cache_middleware)
        self._w3.middleware_onion.add(middleware.latest_block_based_cache_middleware)
        self._w3.middleware_onion.add(middleware.simple_cache_middleware)

        logger.info('Instanced %s client: %s', self.__class__.__name__, self._w3.clientVersion)

        self._chain_id = chain_id

    def add_contract_symbol(self, contract_address: str, symbol: str):
        self.contract_symbols[Web3.toChecksumAddress(contract_address)] = symbol.upper()

    def last_block_number(self) -> int:
        return self._w3.eth.block_number

    def status(self) -> bool:
        return self._w3.isConnected()

    @staticmethod
    def get_new_private_key() -> str:
        return secrets.token_hex(32)

    @staticmethod
    def get_address_from_privatekey(private_key: str) -> str:
        return Web3.toChecksumAddress(Account.from_key(private_key).address)

    @staticmethod
    def get_address_from_publickey(public_key: str) -> str:
        _h = Web3.keccak(hexstr='%s' % public_key)
        return Web3.toChecksumAddress("0x%s" % _h.hex()[-40:])

    def get_contract(self, contract_address: str, _abi: list):
        contract = self._w3.eth.contract(
            Web3.toChecksumAddress(contract_address), abi=_abi)

        return contract

    def get_balance(self, address: str, contract_address: str = None, _abi: str = None) -> int:
        if contract_address is None:
            # Return native token balance
            return self._w3.eth.get_balance(Web3.toChecksumAddress(address))

        # Call the balanceOf function of the smart contract
        contract = self._w3.eth.contract(
            Web3.toChecksumAddress(contract_address),
            abi=self.default_abi.get('balanceOf', None) if _abi is None else _abi)

        return contract.functions.balanceOf(
            Web3.toChecksumAddress(address)).call()

    def get_block(self, block_identifier: int) -> Block:
        _block = self._w3.eth.get_block(block_identifier)

        return mapper.to(Block).map(_block, use_deepcopy=False, fields_mapping={
            "transaction_count": len(_block.transactions)
        })

    @cached(is_contract_cache)
    def is_contract(self, address: str) -> bool:
        """The best way to detect if address is a contract or a wallet is to check if it has
        a contract bytecode, while this call is slow, caching is needed."""
        if self._w3.eth.get_code(address) != b'':
            return True
        else:
            return False

    def get_transfer(
            self, tx_hash: bytearray = None, transaction: dict = None) -> Union[Transfer, None]:
        if transaction is None and tx_hash is None:
            # At least, one of the args must be provided
            return None

        if tx_hash is not None and transaction is None:
            try:
                # Get transaction from chain
                transaction = self._w3.eth.get_transaction(tx_hash)
            except TransactionNotFound:
                return None

        if 'to' not in transaction or transaction.to is None:
            # A transaction without a 'to' address cannot be a transfer
            return None

        # Convert transaction to Transfer
        transfer = None
        # Checking if the receiver is a contract address ?
        contract_symbol = self.contract_symbols.get(Web3.toChecksumAddress(transaction.to), None)
        # Calling the is_contract() is expensive (even it's cached), it is now replaced
        # by simply checking if it's a contract we know (provisioned in self.contract_symbols)
        # all other transactions will be considered as native token transactions, this is
        # acceptable when we're looking for transactions from/to -WATCHED- addresses only.
        #if contract_symbol is None and not self.is_contract(transaction.to):
        if contract_symbol is None:
            transfer = mapper.to(Transfer).map(transaction, use_deepcopy=False, fields_mapping={
                "trx_index": transaction.transactionIndex,
                "hash": HexBytes(transaction.hash),
                "method_id": 'Transfer',
                "sender": Web3.toChecksumAddress(transaction.get('from')),
                "receiver": Web3.toChecksumAddress(transaction.to),
                "block_number": transaction.blockNumber,
                "block_hash": HexBytes(transaction.blockHash),
                "value": transaction['value'],
                "symbol": self.native_token,
                "is_token": False,
                "status": None
            })
        elif contract_symbol is not None:
            # We have a smart contract here
            _tx_data = transaction['input']
            transfer_sig = self.transfer_sigs.get(_tx_data[2:][:8], None)

            # Is it a transfer ?
            if transfer_sig is not None:
                function_args = transfer_sig[1:]

                # Get transaction inputs
                inputs = abi.decode(
                    [_ for _ in function_args], bytearray.fromhex(_tx_data[10:]))

                transfer = mapper.to(Transfer).map(transaction, use_deepcopy=False, fields_mapping={
                    "trx_index": transaction.transactionIndex,
                    "hash": HexBytes(transaction.hash),
                    "method_id": '%s(%s)' % (transfer_sig[0], [_ for _ in function_args]),
                    "sender": Web3.toChecksumAddress(transaction.get('from')),
                    "receiver": Web3.toChecksumAddress(inputs[0]),
                    "block_number": transaction.blockNumber,
                    "block_hash": HexBytes(transaction.blockHash),
                    "value": inputs[1],
                    "symbol": contract_symbol,
                    "is_token": True,
                    "status": None
                })

        return transfer

    def get_transfers(self, block_identifier: int) -> Sequence[Transfer]:
        _transactions = self._w3.eth.get_block(block_identifier, full_transactions=True).get('transactions', [])

        transfers = []
        for _transaction in _transactions:
            transfer = self.get_transfer(transaction=_transaction)

            # Validations
            if transfer is None:
                continue
            if transfer.receiver is None:
                logger.error("Ignoring transfer %s in block #%s: undefined 'to' key",
                             transfer.trx_index, block_identifier)
                continue
            if transfer.receiver == transfer.sender:
                logger.warning("Ignoring transfer %s in block #%s: from == to",
                               transfer.trx_index, block_identifier)
                continue

            transfers.append(transfer)

        return transfers

    def get_transaction_receipt(self, tx_hash: str) -> Union[Receipt, None]:
        try:
            # Get transaction's receipt info from chain
            _receipt = self._w3.eth.get_transaction_receipt(tx_hash)
        except TransactionNotFound:
            return None

        return mapper.to(Receipt).map(_receipt, use_deepcopy=False, fields_mapping={
            "hash": _receipt.transactionHash,
            "status": True if _receipt.get('status', 0) == 1 else False,
            "fee": _receipt.effectiveGasPrice * _receipt.gasUsed,
        })

    def estimate_transact_fee(self, tx: dict) -> tuple[int, int, int]:
        # Gas calculation
        gas_estimate = self._w3.eth.estimate_gas(tx)
        # ref: https://docs.alchemy.com/docs/maxpriorityfeepergas-vs-maxfeepergas#what-is-maxfeepergas-a-hrefand-finally-what-is-max-fee-per-gas-idand-finally-what-is-max-fee-per-gasa
        # Reminder:
        # maxPriorityFeePerGas = Spent for miner tips
        # maxFeePerGas = Spent in total per unit of gas
        base_fee_per_gas = int(self._w3.eth.get_block('latest').baseFeePerGas)
        # max_priority_fee_per_gas = int(self._w3.eth.max_priority_fee)
        max_priority_fee_per_gas = int(base_fee_per_gas * self.miner_fee_percent)  # miner's percent from base_fee_per_gas
        max_fee_per_gas = base_fee_per_gas + max_priority_fee_per_gas

        # Returns: gas_estimate, max_fee_per_gas, max_priority_fee_per_gas
        return gas_estimate, max_fee_per_gas, max_priority_fee_per_gas

    def send_and_get_receipt(self, signed_tx) -> tuple[HexBytes, Any]:
        # Send it
        sent_tx = self._w3.eth.send_raw_transaction(signed_tx.rawTransaction)

        # Wait for receipt
        tx_receipt = self._w3.eth.wait_for_transaction_receipt(sent_tx)

        return tx_receipt.transactionHash, tx_receipt

    def sign_function(self, function_call: Any, private_key: str):
        account = Account.from_key(private_key)

        # Prepare transaction
        _nonce = self._w3.eth.get_transaction_count(account.address)
        tx = {
            'chainId': self._chain_id,
            'nonce': _nonce,
            'from': account.address,
            'type': 0x2
        }

        # Build transaction with function_call
        tx = function_call.build_transaction(tx)

        # Gas calculation
        # Based on:
        # https://ethereum.stackexchange.com/questions/128536/how-do-node-providers-e-g-infura-alchemy-determine-gas-pricing-on-contract-w
        last_block = self._w3.eth.get_block('latest')
        del(tx['gasPrice'])
        tx['maxPriorityFeePerGas'] = 1000000000
        tx['maxFeePerGas'] = (last_block['baseFeePerGas'] * 2) + tx['maxPriorityFeePerGas']
        #tx['gas'] = self._w3.eth.estimate_gas(tx)
        #tx['gasPrice'] = self._w3.eth.gas_price

        # sign the transaction
        signed_tx = self._w3.eth.account.sign_transaction(tx, private_key)

        return signed_tx['hash'], signed_tx

    def sign_transact(
            self, private_key: str, to_address: str, amount: int, fees_included: bool = False) -> tuple[bytearray, Any]:
        account = Account.from_key(private_key)

        # Gas calculation
        _gas_price = None
        _gas, _max_fee_per_gas, _max_priority_fee_per_gas = self.estimate_transact_fee({'to': to_address})

        if fees_included:
            # Include the fees in the amount
            # Make a transaction with type 0x1
            _gas_price = self._w3.eth.gas_price
            amount = amount - (_gas * _gas_price)

        # Prepare transaction
        _nonce = self._w3.eth.get_transaction_count(account.address)
        tx = {
            'chainId': self._chain_id,
            'nonce': _nonce,
            'from': account.address,
            'to': Web3.toChecksumAddress(to_address),
            'value': amount,
            'gas': _gas,
        }

        if _gas_price is None:
            tx['type'] = '0x2'
            tx['maxFeePerGas'] = _max_fee_per_gas
            tx['maxPriorityFeePerGas'] = _max_priority_fee_per_gas
        else:
            tx['type'] = '0x1'
            tx['gasPrice'] = _gas_price

        # sign the transaction
        signed_tx = self._w3.eth.account.sign_transaction(tx, private_key)

        return signed_tx['hash'], signed_tx

    def sign_transfer(self, private_key: str, to_address: str, amount: int,
                      contract_address: str, abi: str = None) -> tuple[bytearray, Any]:
        # Prepare contract
        contract = self._w3.eth.contract(
            Web3.toChecksumAddress(contract_address),
            abi=self.default_abi.get('transfer', None) if abi is None else abi)

        function_call = contract.functions.transfer(
            Web3.toChecksumAddress(to_address), amount)

        return self.sign_function(function_call, private_key)



from dataclasses import dataclass
from dataclasses import field


@dataclass
class Receipt(object):
    hash: bytearray = field(metadata={"required": True})

    status: bool = field(metadata={"required": True})
    fee: int = field(metadata={"required": True})



import logging
import codecs
import base58

from typing import Sequence, Union, Any

from hexbytes import HexBytes

from automapper import mapper

from .interface import INode
from Crypto.Hash import keccak
from tronpy import Tron
from tronpy.providers import HTTPProvider
from tronpy.keys import PrivateKey
from tronpy.abi import trx_abi
from tronpy.exceptions import TransactionNotFound, AddressNotFound

from .types import Block, Transfer, Receipt

logger = logging.getLogger('nodes.tron')


def keccak256(data: bytes) -> bytes:
    hasher = keccak.new(digest_bits=256)
    hasher.update(data)
    return hasher.digest()


class TronNode(INode):
    transfer_sigs = {
        'a9059cbb': ['transfer', 'address', 'uint256']
    }
    contract_symbols = {
        'TR7NHqjeKQxGTCi8q8ZY4pL8otSzgjLj6t': 'USDT',
        'TAFjULxiVgT4qWk6UZwjqwZXTSaGaqnVp4': 'BTT'
    }
    native_token = 'TRX'
    default_abi = {
        'transfer': [
            {"outputs": [{"type": "bool"}],
             "inputs": [{"name": "_to", "type": "address"}, {"name": "_value", "type": "uint256"}],
             "name": "transfer", "stateMutability": "Nonpayable", "type": "Function"}],
        'balanceOf': [{"constant": True, "inputs": [{"name": "who", "type": "address"}],
                       "name": "balanceOf", "outputs": [{"name": "", "type": "uint256"}],
                       "payable": False, "stateMutability": "view", "type": "function"}]
    }

    def __init__(self, uri):
        super(TronNode, self).__init__(uri=uri)
        logger.debug('Instantiating TronNode with uri: %s', uri)
        self._tron = Tron(HTTPProvider(self._uri))
        logger.info('Instanced %s client: versionNum=%s',
                    self.__class__.__name__,
                    self._tron.get_node_info().get('configNodeInfo').get('versionNum'))

    def add_contract_symbol(self, contract_address: str, symbol: str):
        self.contract_symbols[contract_address] = symbol.upper()

    def last_block_number(self) -> int:
        return self._tron.get_latest_block_number()

    def status(self):
        try:
            # tronpy does not provide a status, this is a workaround to return a status
            self._tron.get_latest_block_number()
        except Exception as e:
            logger.error('TronNode status() raised error (%s): %s', e.__class__.__name__, e)
            return False
        else:
            return True

    @staticmethod
    def get_new_private_key() -> str:
        return '%s' % PrivateKey.random()

    @staticmethod
    def get_address_from_privatekey(private_key: str) -> str:
        return PrivateKey(bytes.fromhex(private_key)).public_key.to_base58check_address()

    @staticmethod
    def get_address_from_publickey(public_key: str) -> str:
        _h = keccak256(bytes.fromhex(public_key))
        _hex = '41%s' % _h.hex()[-40:]
        return base58.b58encode_check(bytes.fromhex(_hex)).decode('utf-8')

    def get_contract(self, contract_address: str, _abi: list):
        contract = self._tron.get_contract(contract_address)
        contract.abi = _abi

        return contract

    def get_balance(self, address: str, contract_address: str = None, abi: str = None) -> int:
        if contract_address is None:
            try:
                # Return native token balance
                _decimal_balance = self._tron.get_account_balance(address)
                _xbalance = str(_decimal_balance).split('.')
                if len(_xbalance) == 2:
                    # We have decimals
                    # Pad the decimals to 6 (TRX)
                    return int('%s%s' % (_xbalance[0], _xbalance[1].ljust(6, '0')))
                else:
                    return int('%s000000' % _decimal_balance)
            except AddressNotFound:
                return 0

        # Call the balanceOf function of the smart contract
        # Prepare contract
        contract = self.get_contract(
            contract_address,
            self.default_abi.get('balanceOf', None) if abi is None else abi)

        return contract.functions.balanceOf(address)

    def get_bandwidth(self, address: str) -> int:
        _account_resource = self._tron.get_account_resource(address)
        return _account_resource.get('freeNetLimit', 0) - _account_resource.get('freeNetUsed', 0)

    def get_block(self, block_identifier: int) -> Block:
        _block = self._tron.get_block(block_identifier)

        return mapper.to(Block).map(_block, use_deepcopy=False, fields_mapping={
            "number": _block['block_header']['raw_data']['number'],
            "hash": HexBytes(codecs.decode(_block['blockID'], 'hex_codec')),
            "transaction_count": len(_block.get('transactions', []))
        })

    def get_transfer(
            self, tx_hash: bytearray = None, transaction: dict = None) -> Union[Transfer, None]:
        if transaction is None and tx_hash is None:
            # At least, one of the args must be provided
            return None

        if tx_hash is not None and transaction is None:
            try:
                # Get transaction from chain
                transaction = self._tron.get_transaction(tx_hash)
            except TransactionNotFound:
                return None

        _contracts_len = len(transaction['raw_data']['contract'])
        if _contracts_len != 1:
            # Ref: https://stackoverflow.com/questions/64373633/what-is-the-case-where-you-need-multiple-contracts-results-for-a-single-transact
            logger.error("Ignoring transaction %s in block #%s: contracts = %s",
                         transaction['txID'], transaction['block_number'], _contracts_len)
            return None

        # Convert transaction to Transfer
        transfer = None
        _contract = transaction['raw_data']['contract'][0]
        if _contract['type'] == 'TransferContract':
            # Ignoring TransferAssetContract by now since it holds TRC-10 transfers
            # ref: https://github.com/tronprotocol/documentation/blob/master/English_Documentation/TRON_Virtual_Machine/TRC10_TRX_TRANSFER_INTRODUCTION_FOR_EXCHANGES.md

            transfer = mapper.to(Transfer).map(transaction, use_deepcopy=False, fields_mapping={
                "hash": HexBytes(codecs.decode(transaction['txID'], 'hex_codec')),
                "method_id": 'Transfer',
                "sender": _contract['parameter']['value']['owner_address'],
                "receiver": _contract['parameter']['value']['to_address'],
                "value": _contract['parameter']['value']['amount'],
                "symbol": self.native_token,
                "is_token": False,
                "status": transaction['ret'][0]['contractRet']
            })
        elif _contract['type'] == 'TriggerSmartContract' and 'data' in _contract['parameter']['value']:
            _tx_data = _contract['parameter']['value']['data']
            contract_symbol = self.contract_symbols.get(_contract['parameter']['value']['contract_address'], None)
            transfer_sig = self.transfer_sigs.get(_tx_data[:8], None)

            # Is it a transfer ?
            if contract_symbol is not None and transfer_sig is not None:
                function_args = transfer_sig[1:]

                # Workaround to avoid padding issue
                # Ref: https://github.com/ethereum/eth-abi/issues/116
                if function_args[0] == 'address' and _tx_data[8:][:24] == '000000000000000000000041':
                    _tx_data = _tx_data[:8] + '000000000000000000000000' + _tx_data[8:][24:]

                # Get transaction inputs
                inputs = trx_abi.decode_abi(
                    [_ for _ in function_args], bytearray.fromhex(_tx_data[8:]))

                transfer = mapper.to(Transfer).map(transaction, use_deepcopy=False, fields_mapping={
                    "hash": HexBytes(codecs.decode(transaction['txID'], 'hex_codec')),
                    "method_id": '%s(%s)' % (transfer_sig[0], [_ for _ in function_args]),
                    "sender": _contract['parameter']['value']['owner_address'],
                    "receiver": inputs[0],
                    "value": inputs[1],
                    "symbol": contract_symbol,
                    "is_token": True,
                    "status": transaction['ret'][0]['contractRet']
                })

        return transfer

    def get_transfers(self, block_identifier: int) -> Sequence[Transfer]:
        _block = self._tron.get_block(block_identifier)
        _transactions = _block.get('transactions', [])

        transfers = []
        for _transaction in _transactions:
            transfer = self.get_transfer(transaction=_transaction)

            # Validations
            if transfer is None:
                continue
            if transfer.receiver is None:
                logger.error("Ignoring transfer %s in block #%s: undefined 'to' key",
                             transfer.trx_index, block_identifier)
                continue
            if transfer.receiver == transfer.sender:
                logger.warning("Ignoring transfer %s in block #%s: from == to",
                               transfer.trx_index, block_identifier)
                continue

            transfers.append(transfer)

        return transfers

    def get_transaction_receipt(self, tx_hash: str) -> Union[Receipt, None]:
        if tx_hash[:2] == '0x':
            tx_hash = tx_hash[2:]

        try:
            # Get transaction info from chain
            _transaction_info = self._tron.get_transaction_info(tx_hash)
        except TransactionNotFound:
            return None

        _status = True
        if 'result' in _transaction_info and _transaction_info['result'] != 'SUCCESS':
            _status = False

        return mapper.to(Receipt).map(_transaction_info, use_deepcopy=False, fields_mapping={
            "hash": _transaction_info['id'],
            "status": _status,
            "fee": _transaction_info.get('fee', 0),
        })

    def send_and_get_receipt(self, signed_tx) -> tuple[HexBytes, Any]:
        broadcasted_tx = signed_tx.broadcast().wait()

        return HexBytes(codecs.decode(broadcasted_tx['id'], 'hex_codec')), broadcasted_tx

    def sign_function(self, function_call: Any, private_key: str):
        pk = PrivateKey(bytes.fromhex(private_key))

        function_call.sign(pk)

        return bytes.fromhex(function_call.txid), function_call

    def sign_transact(
            self, private_key: str, to_address: str, amount: int, fees_included: bool = False) -> tuple[bytearray, Any]:
        pk = PrivateKey(bytes.fromhex(private_key))

        if fees_included:
            # Include the fees in the amount
            # @TODO: are we going to consume TRX for this transaction or just bandwidth ?
            amount = amount

        tx = (
            self._tron.trx.transfer(
                pk.public_key.to_base58check_address(),
                to_address,
                amount)
            .build()
            .sign(pk)
        )

        return bytes.fromhex(tx.txid), tx

    def sign_transfer(self, private_key: str, to_address: str, amount: int,
                      contract_address: str, abi: str = None) -> tuple[bytearray, Any]:
        pk = PrivateKey(bytes.fromhex(private_key))

        # Prepare contract
        contract = self.get_contract(
            contract_address,
            self.default_abi.get('transfer', None) if abi is None else abi)

        # Base tx
        call_tx = (
            contract.functions.transfer(
                to_address,
                amount)
            .with_owner(pk.public_key.to_base58check_address())
            .build()
        )

        return self.sign_function(call_tx, private_key)

    def freeze(self, private_key: str, amount: int, resource: str = 'ENERGY') -> HexBytes:
        pk = PrivateKey(bytes.fromhex(private_key))

        tx = (
            self._tron.trx.freeze_balance(
                pk.public_key.to_base58check_address(),
                amount,
                resource
            )
            .build()
            .sign(pk)
        )

        logger.debug("Broadcasting freeze trx: %s", tx)
        broadcasted_tx = tx.broadcast().wait()
        logger.info(
            "Freeze %s TRX from %s for %s, hash=%s",
            amount, pk.public_key.to_base58check_address(), resource, broadcasted_tx['id']
        )

        return HexBytes(codecs.decode(broadcasted_tx['id'], 'hex_codec'))



from abc import ABCMeta, abstractmethod
from typing import Sequence, Union, Any

from .types import Block, Transfer, Receipt


class INode(metaclass=ABCMeta):
    transfer_sigs = {}
    contract_symbols = {}
    native_token = None
    default_abi = {}

    def __init__(self, uri: str):
        self._uri = uri

    def reset_contract_symbols(self):
        self.contract_symbols = {}

    @abstractmethod
    def add_contract_symbol(self, contract_address: str, symbol: str):
        """Append to self.contract_symbols"""

    @abstractmethod
    def last_block_number(self) -> int:
        """Get the last block number on chain"""

    @abstractmethod
    def status(self) -> bool:
        """Indicate node status"""

    @staticmethod
    @abstractmethod
    def get_new_private_key() -> str:
        """Generate a new private key"""

    @staticmethod
    @abstractmethod
    def get_address_from_privatekey(private_key: str) -> str:
        """Get address from private_key"""

    @staticmethod
    @abstractmethod
    def get_address_from_publickey(public_key: str) -> str:
        """Get address from public_key"""

    @abstractmethod
    def get_contract(self, contract_address: str, _abi: list):
        """Returns a contract object"""

    @abstractmethod
    def get_balance(self, address: str, contract_address: str = None, abi: str = None) -> int:
        """Return account balance using public address"""

    @abstractmethod
    def get_block(self, block_identifier: int) -> Block:
        """Get block by number"""

    @abstractmethod
    def get_transfer(
            self, tx_hash: bytearray = None, transaction: dict = None) -> Union[Transfer, None]:
        """Get a Transfer object from an already obtained transaction or by looking up using transaction hash"""

    @abstractmethod
    def get_transfers(self, block_identifier: int) -> Sequence[Transfer]:
        """Get Transfer transactions in a block by number"""

    @abstractmethod
    def get_transaction_receipt(self, tx_hash: str) -> Union[Receipt, None]:
        """Return a Receipt Dto of a transaction by hash"""

    @abstractmethod
    def send_and_get_receipt(self, signed_tx) -> bytearray:
        """Send the transaction and return the receipt"""

    @abstractmethod
    def sign_function(self, function_call: Any, private_key: str):
        """Sign a function call"""

    @abstractmethod
    def sign_transact(
            self, private_key: str, to_address: str, amount: int, fees_included: bool = False) -> tuple[bytearray, Any]:
        """Sign a transaction and return transaction hash and transaction object
        If amount is None, then the balance will be transferred totally to to_address"""

    @abstractmethod
    def sign_transfer(self, private_key: str, to_address: str, amount: int,
                      contract_address: str, abi: str = None) -> tuple[bytearray, Any]:
        """Execute smart contract's transfer function and return transaction hash and transaction object"""



import logging
from .bsc import BscNode
from .ether import EtherNode
from .tron import TronNode
from .exceptions import UnknownApi
from .interface import INode

logger = logging.getLogger('nodes.factory')


class NodeFactory(object):
    def build_node(self, node_id: str, api: str, uri: str, provider: str, **kwargs) -> INode:
        logger.info('[%s] Building node %s from %s', api, node_id, provider)

        if api == 'web3.eth':
            return EtherNode(uri=uri, chain_id=kwargs['chain_id'])
        elif api == 'web3.bsc':
            return BscNode(uri=uri, chain_id=kwargs['chain_id'])
        elif api == 'tronpy':
            return TronNode(uri=uri)
        else:
            raise UnknownApi(api)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nodelib', 'nodelib.types']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=5.2.1,<6.0.0',
 'hexbytes>=0.3.0,<0.4.0',
 'py-automapper>=1.2.3,<2.0.0',
 'tronpy>=0.2.6,<0.3.0',
 'web3>=5.31.3,<6.0.0']

setup_kwargs = {
    'name': 'nodelib',
    'version': '0.2.0',
    'description': '',
    'long_description': 'None',
    'author': 'farirat',
    'author_email': 'fourat@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
